# -*- coding: utf-8 -*-
"""
Created on Wed Jul 28 18:27:45 2021

@author: William Scott Raymond
"""

import os
import shutil
import time
import sys
import re
import subprocess
import platform
import hashlib
import numpy as np

class Error(Exception):
    """Base class for exceptions in this module."""
    pass


class EigenMissingError(Error):
    """Exception raised for when an eigen instillation cannot be found

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

class ExistenceError(Error):
    """Exception raised for when a requesting making a model that already
    exists without overwite == True

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

class ModelNameError(Error):
    """Exception raised for when a requesting making a model that already
    exists without overwite == True

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message



class MisMatchedBrackets(Error):
    """Exception raised for when a requesting making a model that already
    exists without overwite == True

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

class MisMatchedParenthesis(Error):
    """Exception raised for when a requesting making a model that already
    exists without overwite == True

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

class GenericMetaData():
    '''
    Class that generates some generic metadata and returns it in dictionary

    meta data currently returned:
        * user
        * id
        * datetime this function was called
        * python version
        * platform this is being run on

    '''

    def __init__(self):
        self.id = ''
        self.created_at = time.strftime('%Y-%m-%d %H:%M:%S',
                                        time.localtime(time.time()))
        self.user = os.path.expanduser("~")
        self.platform = platform.platform()
        self.python_version = sys.version

    def get(self):
        '''
        generate and return a metadata dictionary for a solver object

        Returns
        -------
        dict
            a dictionary of metadata such as solution id, time ran, user,
            platform and rss version.

        '''
        return self.__dict__

class ModelFactory():
    '''
    Model factory class, this class parses, makes, and compiles custom
    Nascent Chain Tracking (NCT) TASEP models for the rsnapsim package

    the goal of this class is to provide the user a quick way to edit only
    the propensity function of a blank TASEP model such that they dont have to
    write and recompile a full c++/cython model themselves
    '''
    def __init__(self,):
        self.file_path = os.path.dirname(os.path.realpath(__file__))
        self.test_prop = '''

                            if (X_states(0,0) == 1){ // if off turn on rate
                                wn(0) = parameters[0];
                                forward_rate_matrix(0,49) = 0;  // turn on the pause secondary structure is present
                            }
                            if (X_states(0,1) == 1){ // if on turn off rate
                                if (X_full.block(0,50,1,10).sum() == 0){
                                    wn(1) = parameters[1];
                                }
                            }


                            int loc_not_free = (((X_spatial.block(0,0,1,n_ribosomes).array())-0).abs() < R).cast<int>().sum() ;
                            if (loc_not_free == 0){  // kin
                                wn(2) = parameters[2];
                            }

                            if (X_full(0,max_length-1) == 1){  // kout
                                forward_rate_matrix(0,max_length-1) = 0;
                                wn(3) = parameters[3];
                            }
                        '''
                    
        try:
            self.eigen_path = self.find_eigen_path()
            #print('eigen instillation found...')
        except:
            self.eigen_path=''
        self.reserved_model_names = ['build', 'model_maker_cpp', 'models',
                                     'rsnapsim_model_maker']
        self.find_models()




    def find_models(self):
        '''
        function to find all models (compiled or failed to compile)

        This function walks the directory and finds all names
        used in ./models/*/

        Returns
        -------
        models : list of str
            strings of all model names found (folders in ./models).

        '''
        fpath = self.file_path
        build_dirs = [y for y in [x[0] for x in os.walk(fpath)] if 'build' in y]
        split_paths = []
        for pathstr in build_dirs:
            path = os.path.normpath(pathstr)
            split_paths = split_paths + [path.split(os.sep), ]
        models = []
        for pathstr in split_paths:
            try:
                if pathstr[-2] == 'build':
                    if pathstr[-3] not in models:
                        models.append(pathstr[-3])
            except:
                pass

        for model in models:
            if model in self.reserved_model_names:
                models.remove(model)
        self.available_models = models

        return models


    def find_eigen_path(self):
        '''
        find an instillation and fpath of eigen in the current python
        enviroment to use to compile models.

        Raises
        ------
        EigenMissingError
            Raised when Eigen is not installed, we cannot compile c++ models.

        Returns
        -------
        eigen_paths : list of str
            list of all file paths to eigen instillations.

        '''

        paths = sys.path
        potential_paths = []  #check the <env>/lib/ folder
        for path in paths:

            if path[-3:] == 'lib':
                potential_paths.append(path)
            if path[-7:] == 'include':
                potential_paths.append(path)
        eigen_paths = []
        for path in potential_paths:  #in each of these try to find an eigen instillation
            base, _ = os.path.split(path)
            if os.path.exists(os.path.join(base, 'Library',
                                           'include', 'eigen3')):

                eigen_paths.append(os.path.join(base, 'Library',
                                                'include', ''))
            if os.path.exists(os.path.join(base, 'Eigen')):
                eigen_paths.append(os.path.join(base,''))
            if os.path.exists(os.path.join(base, 'eigen3')):
                eigen_paths.append(os.path.join(base,''))
                
        if len(eigen_paths) == 0:
            raise EigenMissingError('Eigen is missing, please provide a path'\
                                    ' or if using a conda instillation, use'\
                                        ' conda install eigen')
        return eigen_paths


    def generate_metadata(self, prop_str, model_name):
        '''
        Generate some meta data to write to the __init__.py of the model

        Parameters
        ----------
        prop_str : str
            propensity string used to generate the model.
        model_name : str
            model name being generated.

        Returns
        -------
        metadata : dict
            dictionary of the metadata.
        meta_str : str
            string version of the metadata
            (this is written to the top of the __init.py__).

        '''
        metadata = GenericMetaData().get() #get all the metadata
        # convert the prop string to a hash ID
        metadata['id'] = hashlib.sha256(prop_str.encode('ascii')).hexdigest()
        # convert to a metadata string
        meta_str = ''
        meta_str += 'model name : ' + model_name + '\n'
        meta_str += 'model ID : ' + metadata['id'] + '\n'
        meta_str += 'files created at : ' + metadata['created_at']+ '\n'
        meta_str += 'platform : ' + metadata['platform']+ '\n'
        meta_str += 'python version : ' + metadata['python_version']+ '\n'

        return metadata, meta_str


    def edit_cpp_files(self, model_name, cpp_file_path, prop_str):
        '''
        edit the c++ files to add the parsed propensity strings

        Parameters
        ----------
        model_name : str
            model.
        cpp_file_path : str
            path to the c++ file to edit.
        prop_str : str
            parsed propensity function string.

        Returns
        -------
        None.

        '''
        strfrnt = '// Start of autogenerated propensity function \n'
        strback = '\n// End of autogenerated propensity function \n'
        prop_str = strfrnt + prop_str + strback
        fstr = ''
        with open(cpp_file_path, 'r') as fname:
            for line in fname:
                if "//INSERT_GENERATED_PROPENSITY_HERE" in line:
                    print('replacing_prop...')
                    line = line.replace("//INSERT_GENERATED_PROPENSITY_HERE",
                                        prop_str)

                fstr += line
        with open(cpp_file_path, 'w') as fname:
            fname.write(fstr)

    def edit_pyx_files(self, model_name, min_length, rules_str,
                       original_rules_str, pyx_file_path):
        '''
        edit the pyx files for the model being made

        Parameters
        ----------
        model_name : str
            name of the model being made.
        min_length : int
            minimum index length in the rules strings provided.
        rules_str : str
            parsed rules string (in c++).
        original_rules_str : str
            original rules string that was passed (python-like).
        pyx_file_path : str
            path to the pyx file to edit (model_name.pyx).

        Returns
        -------
        None.

        '''

        #edit each string to edit
        cdef_str = "cdef extern from 'model_%s.h':"%model_name
        rules_str_r = "rules_str = '''%s"%rules_str + "'''"
        original_rules_str_r = "original_rules_str = ''' %s"%original_rules_str + "'''"
        min_length_str = "min_length = %i"%min_length

        #edit all the lines in the .pyx file
        fstr = ''
        with open(pyx_file_path, 'r') as fname:
            for line in fname:

                if "#cdef goes here" in line:
                    line = line.replace("#cdef goes here", cdef_str)
                if "#min length goes here" in line:
                    line = line.replace("#min length goes here",
                                        min_length_str)
                if "#parsed rules go here" in line:
                    line = line.replace("#parsed rules go here",
                                        rules_str_r)
                if "#original rules go here" in line:
                    line = line.replace("#original rules go here",
                                        original_rules_str_r)
                fstr += line

        with open(pyx_file_path, 'w') as fname:
            fname.write(fstr)


    def edit_setup_files(self, model_name, setup_file_path, eigen_path=''):
        '''
        Edit the setup.py files for the model being made

        Parameters
        ----------
        model_name : str
            name of the model being generated.
        setup_file_path : str
            path of the new setup file to rewrite.
        eigen_path : str, optional
            path to eigen instillation, this has to be added
            to the setup include list. The default is ''.

        Returns
        -------
        None.

        '''
        
        if len(eigen_path) == 0:
            eigen_path = self.eigen_path[0]

        #eigen_path_str = r'{}'.format(eigen_path) #convert to raw string

    #generate all the lines that need to be edited in the setup_model_name.py
        #we need to replace a couple things:
            #the source names (.cpp and .pyx)
            #the cythonize command
            #the setup = "model_name"
            #eigen include path
            #model name string
        src_str = "sources = ['%s.pyx', 'model_%s.cpp']"%(model_name,
                                                          model_name)
        cythonize_str = "cythonize('%s.pyx')"%model_name
        setup_str = "setup(name='%s',"%model_name
        eigen_str = r"include_list = include_list + [%s ,]"%repr(eigen_path)
        model_name_str = "model_name = '%s'"%model_name

        #go through and edit each line to edit
        fstr = ''
        with open(setup_file_path, 'r') as fname:

            for line in fname:

                if "#sources_go_here" in line:
                    line = line.replace("#sources_go_here", src_str)
                if "#cythonize_goes_here" in line:
                    line = line.replace("#cythonize_goes_here", cythonize_str)
                if "#setup_goes_here" in line:
                    line = line.replace("#setup_goes_here", setup_str)
                if "#eigen_path_goes_here" in line:
                    line = line.replace("#eigen_path_goes_here", eigen_str)
                if "#model_name_goes_here" in line:
                    line = line.replace('#model_name_goes_here',
                                        model_name_str)
                fstr += line
        with open(setup_file_path, 'w') as fname:
            fname.write(fstr)


    def make_init_file(self, folder_path, model_name, rules):
        '''
        make the init_file for the model maker

        Parameters
        ----------
        folder_path : str
            The location of where these files are being made
            (where the model maker is installed).
        model_name : str
            name of the model being made.
        rules : str
            additional rule string, this is converted to a hash
            for the model id metadata

        Returns
        -------
        None.

        '''
        #make metadata
        _, meta_str = self.generate_metadata(rules, model_name)

        init_str = ''
        init_str += "'''\n"
        init_str += meta_str
        init_str += "'''\n\n"

        #add the model name to the string to write
        init_str += 'from . import %s'%model_name

        #make the new init file
        initpath = os.path.join(folder_path, 'models',
                                model_name, '__init__.py')
        with open(initpath, "w") as fname:
            fname.write(init_str)



    def generate_model_files(self, model_name, overwrite=False,
                             eigen_path='', propensity_str=None,
                             original_rules_str=None, min_length=None):
        '''
        Generate all model files required to compile a new model and put
        them in a new model folder

        Parameters
        ----------
        model_name : str
            name of the model to generate.
        overwrite : bool, optional
            overwrite any existing files in ./models/<model_name>. The default is False.
        eigen_path : str, optional
            location of an eigen instillation. The default is ''.
        propensity_str : str, optional
            parsed additional rules to paste into the c++ files. The default is None.
        original_rules_str : str, optional
            original unparsed additional rules to paste into the c++ files.
            The default is None.
        min_length : int, optional
            the maximum index for X_lattice detected (minimum length for a tasep).
            The default is None.

        Raises
        ------
        ExistenceError
            Model folder already exists and overwrite == False.

        Returns
        -------
        None.

        '''
        setup_name = 'setup_' + model_name
        model_c_file_name = 'model_' + model_name
        model_h_file_name = 'model_' + model_name

       # model_base_found = False
        #setup_file_found = False
        #model_c_file_found = False
        #model_h_file_found = False

        for root, dirs, files in os.walk('.'):
            for fname in files:
                old_name = os.path.join(os.path.abspath(root), fname)
                base, extension = os.path.splitext(fname)

                if extension == '.pyx':
                    if base == 'blank_model':
                        folder_path = os.path.abspath(root)
                        model_file_path = old_name
                        #model_base_found = True

                        new_model_file = os.path.join(os.path.abspath(root),
                                                      model_name + extension)

                if extension == '.py':
                    if base == 'setup_blank':
                        setup_file_path = old_name
                        #setup_file_found = True

                        new_setup_file = os.path.join(os.path.abspath(root),
                                                      setup_name + extension)


                if extension == '.cpp':
                    if base == 'blank_model':
                        model_c_file_path = old_name
                        #model_c_file_found = True
                        new_model_c_file = os.path.join(os.path.abspath(root),
                                                        model_c_file_name + extension)

                if extension == '.h':
                    if base == 'blank_model':
                        model_h_file_path = old_name
                        #model_h_file_found = True
                        new_model_h_file = os.path.join(os.path.abspath(root),
                                                        model_h_file_name + extension)

        files_exist = np.sum(np.array([os.path.exists(new_model_file),
                                       os.path.exists(new_setup_file),
                                       os.path.exists(new_model_c_file),
                                       os.path.exists(new_model_h_file)]))

        folder_exist = os.path.exists(os.path.join(folder_path, 'models',
                                                   model_name))
        if folder_exist == 0:
            os.makedirs(os.path.join(folder_path, 'models', model_name))
        else:
            if overwrite:
                shutil.rmtree(os.path.join(folder_path, 'models', model_name))
                os.makedirs(os.path.join(folder_path, 'models', model_name))
            else:
                message = 'model folder by the name "%s" exists at %s already'\
                    ' exists, if you wish to overwrite '\
                        'use overwrite == True'% (model_name, folder_path)
                raise ExistenceError(message)

        if files_exist == 0:
            shutil.copy(model_file_path, new_model_file)
            shutil.copy(setup_file_path, new_setup_file)
            shutil.copy(model_c_file_path, new_model_c_file)
            shutil.copy(model_h_file_path, new_model_h_file)


        else:
            if overwrite:
                shutil.copy(model_file_path, new_model_file)
                shutil.copy(setup_file_path, new_setup_file)
                shutil.copy(model_c_file_path, new_model_c_file)
                shutil.copy(model_h_file_path, new_model_h_file)


            else:
                message = 'model files: by the name "%s" exists at %s already'\
                    ' exists, if you wish to overwrite '\
                        'use overwrite == True'% (model_name, new_model_file)
                raise ExistenceError(message)
        time.sleep(.1)
        self.edit_setup_files(model_name, new_setup_file,
                              eigen_path=eigen_path)
        self.edit_pyx_files(model_name, min_length, propensity_str,
                            original_rules_str, new_model_file)


        self.edit_cpp_files(model_name, new_model_c_file, propensity_str)
        shutil.move(new_model_file, os.path.join(folder_path,
                                                 'models', model_name))
        shutil.move(new_setup_file, os.path.join(folder_path,
                                                 'models', model_name))
        shutil.move(new_model_c_file, os.path.join(folder_path,
                                                   'models', model_name))
        shutil.move(new_model_h_file, os.path.join(folder_path,
                                                   'models', model_name))
        self.make_init_file(folder_path, model_name, propensity_str)

        #attempt to compile


    def compile_model(self, model_name, rules=None, overwrite=False,
                      eigen_path='', verbose=True):
        '''
        Compile a custom Nascent Chain tracking (NCT) TASEP model.

        This command takes a set of rules and will replace the propensity function
        inside a set of "blank" c++/pyx files, shuttle these files to a new model folder
        and attempt to compile this model.

        The model takes the following:
            ./blank_model.cpp
            ./blank_model.pyx
            ./setup_blank.py
            ./blank_model.h

        and will make a new folder containing:
            ./models/<model_name>/__init__.py
            ./models/<model_name>/model_<model_name>.cpp
            ./models/<model_name>/model_<model_name>.h
            ./models/<model_name>/<model_name>.pyx
            ./models/<model_name>/setup_<model_name>.py

        and will then run "python setup_<model_name>.py build_ext --inplace"
        and attempt to compile the model. If successful the model can then be imported
        via:

            from models import <model_name>

        Parameters
        ----------
        model_name : str
            name of the model to try to compile.
        rules : str, optional
            The additional rules containing the propensity function
            to replace the one in the c++ source. The default is None.
        overwrite : bool, optional
            option to overwrite an existing model, otherwise if the model
            name is already in use, it will not delete or recompile the model
            folder. The default is False.
        eigen_path : str, optional
            location to the path of the eigen instillation
            https://eigen.tuxfamily.org/. The default is ''.
        verbose : bool, optional
            print out the steps of compiling and errors if it goes
            wrong. The default is True.

        Raises
        ------
        ModelNameError
            Raised if a model name is already in use and overwrite == False.

        Returns
        -------
        None.

        '''
        #if the user did not provide a rule string, use a dummy one
        if rules == None:
            rules = self.test_prop
        if model_name in self.reserved_model_names:
            msg = 'The model name requested is a reserved keyword and'\
                ' cannot be used to build the model files, please rename'\
                    ' the model.'
            raise ModelNameError(msg)

        # convert the rules passed
        if verbose:
            print('parsing rules...')
            parsed_rules = RuleConverter().parse_rules(rules)
            min_length = RuleConverter().get_min_length(rules)

        if verbose:
            print('generating model files...')

        #generate all the appropriate files
        cwd = os.getcwd()
        os.chdir(self.file_path) #change to the cwd where this file is stored
        self.generate_model_files(model_name, overwrite=overwrite,
                                  eigen_path=eigen_path,
                                  propensity_str=parsed_rules,
                                  min_length=min_length,
                                  original_rules_str=rules)
        #change dir to where the new files are
        os.chdir(os.path.join('models', model_name))
        #attempt to compile
        if verbose:
            print('compiling model...')

        if verbose:
            compile_run = subprocess.run(["python", "setup_%s.py"%model_name,
                                          "build_ext", "--inplace"],
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE)

        else:
            compile_run = subprocess.run(["python", "setup_%s.py"%model_name,
                                          "build_ext", "--inplace"])

        # check if the model compiled
        if compile_run.returncode != 0:
            if verbose:
                print('compilation failed...')
                print(compile_run.stdout.decode('utf-8'))
                print(compile_run.stderr.decode('utf-8'))
        else:
            if verbose:
                print('model compiled!')
        os.chdir(cwd)




class RuleConverter():
    '''
    Class that converts the python-like rules string into c++ to paste into
    the blank model files



    Full example:

        additional_rules = ```

        #propensity function for hairpin model
        int hairpin_location = cast_to_int(parameters[4]) # location of the hairpin
        if state[0] == 1:
            #if the hairpin formed, state = off
            wn[0] = parameters[0]
            step[ hairpin_location-1  ] = 0

        if state[1] == 1:
            #if the hairpin formed, state = on
            if sum(X[hairpin_location:hairpin_location+10  ]) < 1:
                wn[1] = parameters[1]

        if free[0]:
            #if the front location is free, allow kin
            wn[2] = parameters[2]

        if X[149] == 1:
            # if its in the final location, add kout
            step[149] = 0
            wn[3] = parameters[3]



        ```

        >>>parsed_rules = RuleConverter().parse_rules(additional_rules)


        parsed_rules = ```

            //propensity function for hairpin model
            int hairpin_location = (static_cast<int>(parameters[4] < 0 ? parameters[4] - 0.5 : parameters[4] + 0.5)); // location of the hairpin
            if( X_states(0) == 1){
            //if the hairpin formed, state = off
                wn[0] = parameters[0];
                forward_rate_matrix( hairpin_location-1  ) = 0;
                }

            if( X_states(1) == 1){
            //if the hairpin formed, state = on
                if( X_full.block(0,hairpin_location,1,hairpin_location+10  ).sum() < 1){
                    wn[1] = parameters[1];
                    }
                }

            if(   ((((X_spatial.block(0,0,1,n_ribosomes).array())-0).abs() < R).cast<int>().sum() == 0) ){
            //if the front location is free, allow kin
                wn[2] = parameters[2];
                }

            if( X_full(149) == 1){
            // if its in the final location, add kout
                forward_rate_matrix(149) = 0;
                wn[3] = parameters[3];
                }



        ```




    '''
    def __init__(self):

        #functions denoted by () to find and convert
        self.keyfuns = {'sum': '.sum()',
                        'cast_to_int':'static_cast<int>(%s < 0 ? %s - 0.5 :'\
                                      ' %s + 0.5)'}

        #keywords with [] to find and convert
        self.keywords = {'state': 'X_states',
                         'wn':'wn',
                         'step':'forward_rate_matrix',
                         'X':'X_full',
                         'occupied':'  ((((X_spatial.block(0,0,1,n_ribosomes)'\
                                    '.array())-%s).abs() < R).cast<int>().'\
                                    'sum() != 0) ',
                         'free':'  ((((X_spatial.block(0,0,1,n_ribosomes).array()'\
                                ')-%s).abs() < R).cast<int>().sum() == 0) ',}

        #if operators to parse
        self.if_operators = ['<', '>', '==', '!=', '>=', '<=']


    '''
    these are not ready yet
    def make_in(self,in_loc, reaction_num, par_num):

        in_str = 'int loc_not_free = (((X_spatial.block(0,0,1,n_ribosomes).array())-0).abs() < R).cast<int>().sum() ;\nif (loc_not_free == 0){  // kin\n'


        in_str = in_str + 'wn(%i) = parameters[%i];'%(reaction_num, par_num)
        in_str = in_str + '}'

        return in_str

    def make_out(self, out_loc, reaction_num, par_num, run_through):

        out_str = 'if (X_full(0,max_length-1) == 1){  // kout\n '
        if run_through:
            out_str = 'forward_rate_matrix(0,max_length-1) = 0;\n'

        out_str = out_str + 'wn(%i) = parameters[%i];'%(reaction_num, par_num)
        out_str = out_str + '}'

        return out_str

    '''

    @staticmethod
    def check_brackets(string):

        '''
        Check if there are mismatched parenthesis in a string,
        returns none if it doesnt error out.

        Parameters
        ----------
        string : str
            string to check.

        Raises
        ------
        MisMatchedParenthesis
            Error if mismatched (.
        MisMatchedBrackets
            Error if mismatched [.

        Returns
        -------
        None.

        '''
        for bracket in ['[', '(']:
            restr = {'[': '\[|\]', '(':'\(|\)'}[bracket]
            par_locations = [(m.start(0), m.end(0)) for m in re.finditer(restr, string)]
            #exception here if unclosed
            total_pars = int(len(par_locations))
            if total_pars%2 != 0:
                if bracket == '(':
                    raise MisMatchedParenthesis('MisMatched Parenthes'/
                                                'is () in the stri'/
                                                'ng: %s'% string)
                if bracket == '[':
                    raise  MisMatchedBrackets('MisMatched Brackets [] '/
                                              'in the string: %s'% string)


    @staticmethod
    def get_parenthesis_dict(string):
        '''
        Get all matching pairs of parenthesis from a string, error if mismatched

        Parameters
        ----------
        string : str
        string to find the parenthesis pairs of

        Raises
        ------
        MisMatchedParenthesis
        Error if there is an unbalanced ( or )

        Returns
        -------
        list of tuples
        list of tuples of matches from a string. Example:
        'hello(3) + hello(5)' would return
        [(5,7), (16,18)]

        '''
        istart = []
        bracket_dict = {}

        for i, char in enumerate(string):
            if char == '(':
                istart.append(i+1)
            if char == ')':
                try:
                    bracket_dict[istart.pop()] = i
                except:
                    raise MisMatchedParenthesis('Missing "("  with closed '\
                                                ' parenthesis at %i'%i)

        if len(istart) != 0:  # still a ( that cant be matched
            raise MisMatchedParenthesis('Mismatched ")" ')
        return list(bracket_dict.items())

    @staticmethod
    def get_square_bracket_dict(string):
        '''
        Get all matching pairs of square brackets from a string, error if mismatched

        Parameters
        ----------
        string : str
        string to find the square bracket pairs of

        Raises
        ------
        MisMatchedBrackets
        Error if there is an unbalanced [ or ]

        Returns
        -------
        list of tuples
        list of tuples of matches from a string. Example:
        'hello[3] + hello[5]' would return
        [(5,7), (16,18)]

        '''
        istart = []
        bracket_dict = {}

        for i, char in enumerate(string):
            if char == '[':
                istart.append(i+1)
            if char == ']':
                try:
                    bracket_dict[istart.pop()] = i
                except:
                    raise MisMatchedBrackets('Missing "["  with closed '\
                                             'parenthesis at %i'%i)

        if len(istart) != 0:  # still a ( that cant be matched
            raise MisMatchedBrackets('Mismatched "]" ')
        return list(bracket_dict.items())

    @staticmethod
    def invert_pair_locations(pair_locations, length):
        '''
        invert a set of pair locations of a string so we have all indexes of all substrings

        example:

            s = 'abc[defg]hijk'
            pair_locations = [(3,8), ] #pair strings of [*]
            length = 12

            result: [(0,3),(8,11) ] #inverted pair strings!

        Parameters
        ----------
        pair_locations : list of tuples of ints
            substring pair locations of brackets in a string in a format [(x,y), (z,w)...] .
        length : int
            total length of the string the pair locations come from.

        Returns
        -------
        inverse_pair_locations : list of tuples of ints
            inverted substring pair locations of brackets (indexes of substrings outside brackets).

        '''
        order = ''
        inverse_pair_locations = []

        if pair_locations[0][0] == 0:
            order += 'm'
        else:
            order += 'i'
            inverse_pair_locations = inverse_pair_locations + [(0, pair_locations[0][0]),]
        for i in range(len(pair_locations)-1):
            inverse_pair_locations = inverse_pair_locations + [(pair_locations[i][1], pair_locations[i+1][0]),]

        if pair_locations[-1][1] != length:
            inverse_pair_locations = inverse_pair_locations + [(pair_locations[-1][1], length),]
            order += 'i'
        else:
            order += 'm'
        return inverse_pair_locations

    @staticmethod
    def get_orders_of_substrings(pair_locations, inverse_pair_locations,):
        '''
        given all pair locations inside and outside of brackets, get the order
        they should be reconstructed in.

        example:

         s = 'abc[defg]hijk'
         pair_locations = [(3,8), ] #pair strings of [*]
         length = 12
         inverse_locations = [(0,3),(8,11)]

         result: [(0,3),(3,8),(8,11)],  'imi'

         imi for inverse, match, inverse

        Parameters
        ----------
        pair_locations : list of tuples of ints
            substring pair locations of brackets in a string in a format [(x,y), (z,w)...].
        inverse_pair_locations : list of tuples of ints
            inverted substring pair locations of brackets (indexes of substrings outside brackets).

        Returns
        -------
        combined_strings : list of tuples of ints
            all pair locations.
        orders : str
            string denoting which pairs are outside ('i') and inside brackets ('m').

        '''
        n_substrings = len(pair_locations) + len(inverse_pair_locations)
        k = 0
        m = 0
        n = 0
        orders = ''
        combined_strings = []
        for i in range(n_substrings):
            if m < len(pair_locations):
                if pair_locations[m][0] == k:
                    k = pair_locations[m][1]
                    combined_strings = combined_strings + [pair_locations[m],]
                    orders += 'm'
                    m += 1
            if n < len(inverse_pair_locations):
                if inverse_pair_locations[n][0] == k:
                    k = inverse_pair_locations[n][1]
                    combined_strings = combined_strings + [inverse_pair_locations[n],]
                    orders += 'i'
                    n += 1
        return combined_strings, orders


    def get_min_length(self, rule_string):
        '''
        Function that gets the minimum length construct needed for a model.

        Parameters
        ----------
        rule_string : str
            The full additional rules string

        Returns
        -------
        int
            Maximum value of an index for the stepping matrix or
            X_lattice matrix. This corresponds to the minimum length
            construct needed by the model when passing values to it.

        '''

        #remove all comments, trailing_whitespace / leading whitespace
        rules_split, _, _, _, _, _, _ = self.get_valid_rules(rule_string)
        all_indexes = []
        for i in range(len(rules_split)):
            a = [(m.start(0), m.end(0)) for m in re.finditer('step\[(.*?)\]',
                                                             rules_split[i])]
            b = [(m.start(0), m.end(0)) for m in re.finditer('X\[(.*?)\]',
                                                             rules_split[i])]

            a = [re.findall(r'\d+', rules_split[i][x[0]:x[1]]) for x in a]
            b = [re.findall(r'\d+', rules_split[i][x[0]:x[1]]) for x in b]
            all_indexes = all_indexes + a + b
        all_indexes = [item for sublist in all_indexes for item in sublist]
        return max([int(x) for x in all_indexes])


    def get_valid_rules(self, rule_string):
        '''
        master function that parses out stand alone comments and
        trailing comment along with pairs of line orders.

        Parameters
        ----------
        rule_string : str
            the string to parse.

        Returns
        -------
        rules_split : list of str
            list of split up lines without trailing comments.
        rules_line_ids : list of ints
            ids of the order of the lines that were rules.
        comments : list of str
            list of stand alone comments (denoted by #).
        comment_line_ids : list of ints
            ids of the order of comments.
        trailing_comments : list of strs
            list of trailing comment strings.
        trailing_comment_ids : list of ints
            ids of the lines containing trailing comments.
        total_lines : int
            N of total lines.

        '''
        rules_split_all = [x for x in rule_string.split('\n') if len(x.lstrip()) != 0]
        total_lines = len(rules_split_all)
        rules_line_ids = [x[0] for x in enumerate(rules_split_all)]

        comments = [x for x in rules_split_all if x.lstrip()[0] == '#']
        comment_line_ids = [x[0] for x in enumerate(rules_split_all) if x[1].lstrip()[0] == '#']

        trailing_comments = []
        trailing_comment_ids = []
        for i in range(len(rules_split_all)):
            if i not in comment_line_ids:
                if '#' in rules_split_all[i]:
                    trailing_comments.append(rules_split_all[i][rules_split_all[i].find('#'):])
                    trailing_comment_ids = trailing_comment_ids + [i,]

        for idx in comment_line_ids:
            rules_line_ids.remove(idx)
        rules_split = []
        for idx in rules_line_ids:
            if idx in trailing_comment_ids:
                rules_split = rules_split + [rules_split_all[idx][:rules_split_all[idx].find('#')],]
            else:
                rules_split = rules_split + [rules_split_all[idx],]

        return rules_split, rules_line_ids, comments, comment_line_ids, trailing_comments, trailing_comment_ids, total_lines

    def parse_rules(self, rule_string):
        '''
        Master function that attempts to convert the given rule set to a
        parsed ruleset

        Parameters
        ----------
        rule_string : str
            the rule string to convert to c++.

        Returns
        -------
        parsed_rule_full_str : str
            full rule string parsed to c++.

        '''

        #split up the rules by lines, remove stand alone line comments
        #(lines starting with #) and trailing commments, (anything after a #)
        rules_split, rules_line_ids, comments, comment_line_ids, trailing_comments, trailing_comment_ids, total_lines = self.get_valid_rules(rule_string)

        indents = [len(x) - len(x.lstrip()) for x in rules_split] #get the indent levels

        #outer levels (where indents = 0)
        #inner levels (where indents != 0)
        outer_levels = np.where(np.array(indents) == 0)[0]
        inner_levels = np.where(np.array(indents) != 0)[0]

        #convert all of these to consistent levels and parse the tabs
        max_outer = np.max(outer_levels)
        max_inner = np.max(inner_levels)
        max_total = np.max([max_outer, max_inner])
        outer_levels = np.append(outer_levels, max_total + 1)

        levels = []
        for i in range(0, len(outer_levels)-1):
            tosort = [indents[x] for x in range(outer_levels[i]+1, outer_levels[i+1])]
            orderkey = list(set(tosort))
            orderkey.sort()
            levels.append([0,])
            levels.append([orderkey.index(tosort[x]) + 1 for x in range(len(tosort))])

        levels.append([0,])

        #strip out all whitespace now that we have the indentation levels
        levels = [item for sublist in levels for item in sublist]
        rules_stripped = [x.lstrip().rstrip() for x in rules_split]
        #parse all the comments (just convert the front # to a //)
        comments_stripped = [x.lstrip().rstrip() for x in comments]
        trailing_comments_stripped = [x.lstrip().rstrip() for x in trailing_comments]
        comments_parsed = ['//' + x[1:] for x in comments_stripped]
        trailing_comments_parsed = ['//' + x[1:] for x in trailing_comments_stripped]

        #now for each valid rule line, convert those as well
        rules_parsed = []
        for i in range(len(rules_stripped)):
            #check if its an if statement to parse
            if rules_stripped[i][:2] == 'if':
                rules_parsed = rules_parsed + [self.convert_funs(
                    self.convert_if(rules_stripped[i])),]

            elif rules_stripped[i][:4] == 'for ':
                rules_parsed = rules_parsed + [self.convert_funs(
                    self.convert_for(rules_stripped[i])),]
            else:
                rules_parsed = rules_parsed + [self.convert_funs(
                    self.convert_assignment(rules_stripped[i])),]

        #step through based on the indentation levels and add the appropriate
        #{} for the c++
        parsed_rule_full_str = ''
        current_level = 0
        for i in range(total_lines):

            if i in rules_line_ids:
                j = rules_line_ids.index(i)

                front_str = '    '*levels[j]
                current_level = levels[j]
                if i in trailing_comment_ids:
                    m = trailing_comment_ids.index(i)
                    parsed_rule_full_str += front_str +  rules_parsed[j] + ' ' + trailing_comments_parsed[m] +  '\n'
                else:
                    parsed_rule_full_str += front_str +  rules_parsed[j] + '\n'
                add_newline = False
                n = levels[j] -levels[j+1] + 1
                for k in range(0, levels[j] -levels[j+1]):
                    add_newline = True
                    n -= 1
                    parsed_rule_full_str += '    '*n + '}\n'

                if add_newline:
                    parsed_rule_full_str += '\n'

            if i in comment_line_ids:
                j = comment_line_ids.index(i)
                front_str = '    '*current_level
                parsed_rule_full_str += front_str +  comments_parsed[j] + '\n'

        #return the final parsed c++ rules
        return parsed_rule_full_str



    def convert_var(self, var, key):
        '''
        convert a keyword variable to its c++ counterpart

        for example:

        X[0:34] will be converted to X_full.block(0,0,1,34)

        Parameters
        ----------
        var : str
        full string to convert.
        key : str
        the python-like key to convert.

        Returns
        -------
        newvar : str
        the variable string converted to c++.


        '''
        newvar = var[:]

        index_str = newvar[len(key)+1:-1] #remove key and brackets

        if key in ['state', 'step', 'X']:
            if ':' not in index_str:
                new_indexes = '(' + index_str +')'

            else:
                j, k = str(index_str.split(':')[0]), str(index_str.split(':')[1])
                new_indexes = '.block(0,%s,1,%s)'%(j, k)

            newvar = self.keywords[key] + new_indexes

        if key in ['free', 'occupied']:
            newvar = self.keywords[key]%index_str

        return newvar


    def convert_keyfuns(self, string, bracket='('):
        '''
        convert key funs (sets bracket to '(' for keystring func )
        '''
        key_list = list(self.keyfuns.keys())
        return self.convert_keystrings(key_list, string, bracket, fun=True)

    def convert_keywords(self, string, bracket='['):
        '''
        convert key words (sets bracket to '[' for keystring func )
        '''
        key_list = list(self.keywords.keys())
        return self.convert_keystrings(key_list, string, bracket)


    def convert_range(self, string):
        '''
        gets the substrings x, y, z from "range(x,y,z)"

        Parameters
        ----------
        string : str
            string to convert to a for loop in c++.

        Returns
        -------
        parsed substrings.

        '''
        strcopy = string[:]
        iters = strcopy[6:-1].split(',')
        if len(iters) == 2:
            range_iter = '1'
            range_start, range_stop = iters
        else:
            range_start, range_stop, range_iter = iters

        return range_start, range_stop, range_iter

    def convert_for(self, string):
        '''
        convert the statement in the format

            "for i in range(x,y,z):"

        to

            for(int i = x; i < y; i+=z){

        Parameters
        ----------
        string : str
            for loop start to convert.

        Returns
        -------
        forstr : str
            for loop statement converted to c++.

        '''
        key = 'range'
        self.check_brackets(string)
        #get pair locations
        pair_locations = self.get_parenthesis_dict(string)
        statements_to_edit = [string[x[0]-len(key)-1:x[1]+1] for x in pair_locations]

        statements_to_edit2 = []
        bracket_to_edit = []
        for i in range(len(statements_to_edit)):

            if key+'(' == statements_to_edit[i][:len(key)+1]:

                statements_to_edit2 = statements_to_edit2 + [statements_to_edit[i],]
                bracket_to_edit = bracket_to_edit + [(pair_locations[i][0] -len(key)-1, pair_locations[i][1]+1),]

            inverse_brackets = self.invert_pair_locations(bracket_to_edit, len(string))
            #get the order of these substring pairs
            substring_pairs, _ = self.get_orders_of_substrings(bracket_to_edit, inverse_brackets)

            #loop through these pairs and rebuild the string, converting the key + bracket function as appropriate
            all_substrings = [string[x[0]:x[1]] for x in  substring_pairs]
            for i in range(len(all_substrings)):

                if key + '(' == all_substrings[i][:len(key)+1]:
                    substring1, substring2, substring3 = self.convert_range(all_substrings[i])


        itername = string[4:string.index(' in ')]
        forstr = 'for(int %s = %s; %s < %s; %s+= %s){'%(itername,
                                                        substring1,
                                                        itername,
                                                        substring2,
                                                        itername,
                                                        substring3)

        return forstr

    def convert_keystrings(self, key_list, string, bracket, fun=False):
        '''
        take any string that has a key word (X, free, occupied) or
        a key function (sum, cast_to_int) + a desired bracket, and then
        convert them to their c++ equivelents and return the string.

        Parameters
        ----------
        key_list : list
        list of str keys to check for.
        string : str
        original string to convert.
        bracket : str
        which bracket to find and convert, ( or [.
        fun : bool, optional
        are we parsing functions? if so we will use edit_statement.
        The default is False.

        Returns
        -------
        string : str
        The converted string.

        '''


        for key in key_list:

            if key + bracket in string:

                self.check_brackets(string)  #check if the brackets are matching
                #if they match get the matching pairs to split by
                if bracket == '[':
                    pair_locations = self.get_square_bracket_dict(string)
                if bracket == '(':
                    pair_locations = self.get_parenthesis_dict(string)

                #split out the statements split by the desired bracket
                statements_to_edit = [string[x[0]-len(key)-1:x[1]+1] for x in pair_locations]

                # find which of these statements contain key + bracket as the start
                statements_to_edit2 = []
                bracket_to_edit = []
                for i in range(len(statements_to_edit)):

                    if key+bracket == statements_to_edit[i][:len(key)+1]:

                        statements_to_edit2 = statements_to_edit2 + [statements_to_edit[i],]
                        bracket_to_edit = bracket_to_edit + [(pair_locations[i][0] -len(key)-1, pair_locations[i][1]+1),]

                #for each of those matching key + bracket, invert the pair locations
                inverse_brackets = self.invert_pair_locations(bracket_to_edit, len(string))
                #get the order of these substring pairs
                substring_pairs, _ = self.get_orders_of_substrings(bracket_to_edit, inverse_brackets)

                #loop through these pairs and rebuild the string, converting the key + bracket function as appropriate
                all_substrings = [string[x[0]:x[1]] for x in  substring_pairs]
                string_to_rebuild = []
                for i in range(len(all_substrings)):
                    #split by the statement
                    if key + bracket == all_substrings[i][:len(key)+1]:
                        if fun:
                            string_to_rebuild = string_to_rebuild + [self.edit_statement(all_substrings[i], key),]
                        else:

                            string_to_rebuild = string_to_rebuild + [self.convert_var(all_substrings[i], key),]
                    else:
                        string_to_rebuild = string_to_rebuild + [all_substrings[i], ]

                string = ''.join(string_to_rebuild) #rebuild the full string
        return string

    def convert_assignment(self, statement_to_parse):
        '''
        convert an assignment such as

        state[0] = 1 to X_state(0) = 1;


        Parameters
        ----------
        statement_to_parse : str
            string to convert.

        Returns
        -------
        statement_copy : TYPE
            the converted assignment statement.

        '''

        statement_copy = statement_to_parse[:] #copy input string
        statement_copy = self.convert_keywords(statement_copy, bracket='[')

        if statement_copy[-1] != ';':
            statement_copy = statement_copy + ';'

        return statement_copy



    def convert_if(self, statement_to_parse):
        '''
        convert an if statement to c++

        Parameters
        ----------
        statement_to_parse : str
            if statment to convert.

        Returns
        -------
        statement_copy : str
            if statement in c++ (before function conversions).

        '''

        statement_copy = statement_to_parse[:] #copy input string
        statement_copy = statement_copy.replace(' and ', ' && ')
        statement_copy = statement_copy.replace(' or ', ' || ')



        statement_copy = self.convert_keywords(statement_copy)
        found_operator = False
        for operator in ['||', '&&']:
            found_operator =  True
            statement_copy = statement_copy.replace(operator,
                                                    ')' + operator + '(')

        statement_copy = 'if(' + statement_copy[2:]
        if statement_copy[-1] == ':':
            statement_copy = statement_copy[:-1]
        statement_copy = statement_copy + '){'
        #we found an operator so now we have to add an extra set of ()
        if found_operator:
            statement_copy = statement_copy.replace('if(', 'if((')
            statement_copy = statement_copy.replace('){', ')){')

        return statement_copy


    def convert_funs(self, statement_to_parse):
        '''
        convert all functions within a string

        Parameters
        ----------
        statement_to_parse : str
            statement string to convert.

        Returns
        -------
        statement_copy : str
            function converted string.

        '''
        statement_copy = statement_to_parse[:]
        statement_copy = self.convert_keyfuns(statement_copy)
        return statement_copy


    def edit_statement(self, string, key):
        '''
        convert the key functions to their c++ equivelant

        for example:

            sum(X_full.block(0,0,1,34)) to X_full.block(0,0,1,34).sum()

        Parameters
        ----------
        string : str
            the string to convert.
        key : str
            key to convert.

        Returns
        -------
        TYPE
            converted statement string.

        '''
        if key == 'sum':

            if string != 'sum()':
                string = string.replace('sum(', '')[:-1] + '.sum()'

        if key == 'cast_to_int':
            string = string.replace('cast_to_int(', '')[:-1]
            string = '(' + self.keyfuns['cast_to_int']%(string, string, string) + ')'
        return string




try:
    from setuptools import find_packages
    from setuptools import setup
    from setuptools import Extension
    
except ImportError:
    from distutils.core import setup
    from distutils.extension import Extension
    
from Cython.Distutils import build_ext
from Cython.Build import cythonize

import sysconfig
import os

# Call cythonize in advance so a single module can be compiled from a single Cython
# file along with other C++ files.
import numpy as np

#sources_go_here
#cythonize_goes_here

if not sysconfig.get_config_var('LIBS'):
    libs = sysconfig.get_config_var('LIBS')
else:
    libs = []


class DependencyError(Exception):
    pass

try:
    packages=find_packages()
except:
    pass

try:
    eca = sysconfig.get_config_var('CPPFLAGS').split()
except:
    eca = []
    
# Try to find eigen directory if possible
include_list = [sysconfig.get_paths()['include'],np.get_include(),'.', os.getcwd()]

#eigen_path_goes_here
#model_name_goes_here
#setup_goes_here
      ext_modules=[Extension(model_name, 
                sources, language='c++',
                include_dirs = include_list ,
                library_dirs = libs,
                extra_compile_args= eca)]
    ,cmdclass = {'build_ext': build_ext}
    ,author='rsnapsim'
    ,description= 'auto_generated_model_cpp_lib'
    ,version = "0.0a"
    ,long_description = 'This file was autogenerated for an rSNAPsim c++ model '
    ,long_description_content_type='text/markdown'
    ,url = ''
    ,install_requires = ['numpy',"Cython>=0.26.0"]
    ,packages=packages
    
    )





# -*- coding: utf-8 -*-
"""
Created on Thu Dec 17 17:50:16 2020

@author: willi
"""

import numpy as np

class GenericSSA():

    '''
    Generic SSA solver - used for the simulated cell animations
    '''

    def __init__(self,type='linear'):

        self.time_variant = False
        self.xi=np.array([])
        self.ti= None
        self.tf=None
        self.S=np.array([])
        self.type=type
        self.ptimes=100
        self.params={}
        if type=='linear':
            #self.fast_rxn = 0.5
            self.W0=np.array([])
            self.W1=np.array([])
        if type == 'nonlinear':
            #self.fast_rxn = 0.5
            self.P=lambda x,t:None



    def gettvec(self):
        return np.linspace(self.ti,self.tf,self.ptimes)

    def updateparams(self):
        for k, v in self.params.iteritems():
            setattr(self, k, v)

    def _run_trajectory(self):#!!!!!!!!!!!!!!!!!renamed run to solve(big deal)
        x=self.xi
        t=self.ti
        __n=len(x)
        self.time=self.gettvec()
        data=np.zeros((len(self.xi),self.ptimes))
        ip=0

        if self.type=='linear':
            if self.time_variant == False:
                while t<self.tf:
                    rate=np.atleast_2d(np.dot(self.W1,x))+self.W0
                    rate=np.cumsum(rate)
                    with np.errstate(divide='ignore', invalid='ignore'):
                        t=(t-np.log(np.random.rand(1))/rate[-1])
                    ro=rate[-1]*np.random.rand()
                    while t>self.time[ip]:
                        if t>self.tf:
                            b = len(self.time[ip:])
                            fill = np.repeat(x,b)
                            data[:,ip:]=fill.reshape(__n,b)
                            return data
                        else:
                            data[:,ip]=x.reshape(__n)
                            ip=ip+1
                    for i in range(len(rate)):
                        if rate[i]>=ro:
                            event=i
                            break
                    x=x+np.atleast_2d(self.S[:,event]).T

            else:


                x = np.concatenate((np.array([0]),self.xi.flatten()))

                #x = self.xi
                __n=len(x)
                self.time=self.gettvec()
                data=np.zeros((len(self.xi),self.ptimes))
                a,b = self.S.shape
                S = np.vstack((np.zeros(b),self.S))
                S = np.hstack((np.zeros((a+1,1)),S))
                while t<self.tf:
                    __n=len(x)
                    self.time=self.gettvec()
                    data=np.zeros((len(self.xi),self.ptimes))
                    a,b = self.S.shape
                    S = np.vstack((np.zeros(b),self.S))
                    S = np.hstack((np.zeros((a+1,1)),S))
                    while t<self.tf:
                        trate=self.get_P(x[1:],t)

                        rate = np.concatenate((np.array([self.fast_rxn]),trate))
                        rate=np.cumsum(rate)


                        t=(t-np.log(np.random.rand(1))/rate[-1])
                        ro=rate[-1]*np.random.rand()

                        while t>self.time[ip]:
                            if t>self.tf:
                                b = len(self.time[ip:])
                                fill = np.repeat(x[1:],b)
                                data[:,ip:]=fill.reshape(__n-1,b)
                                return data
                            else:
                                #data[:,ip]=x.reshape(__n)
                                data[:,ip]=x[1:]
                                ip=ip+1
                        for i in range(len(rate)):
                            if rate[i]>=ro:
                                event=i

                                break

                        x=x+S[:,event].ravel()
                    '''
                    rate=np.atleast_2d(np.dot(self.W1(t),x))+self.W0(t)

                    rate=np.cumsum(rate)

                    t=(t-np.log(np.random.rand(1))/rate[-1])
                    print(t)
                    ro=rate[-1]*np.random.rand()
                    while t>self.time[ip]:
                        if t>self.tf:
                            b = len(self.time[ip:])
                            fill = np.repeat(x,b)
                            data[:,ip:]=fill.reshape(__n,b)
                            return data
                        else:
                            data[:,ip]=x.reshape(__n)
                            ip=ip+1
                    for i in range(len(rate)):
                        if rate[i]>=ro:
                            event=i
                            break
                    x=x+np.atleast_2d(self.S[:,event]).T

                    '''



        elif self.type=='nonlinear':
            if self.time_variant == True:  #if time variant use fast reaction
                x = np.concatenate((np.array([0]),self.xi.flatten()))

                #x = self.xi
                __n=len(x)
                self.time = self.gettvec()
                data = np.zeros((len(self.xi), self.ptimes))
                a, b = self.S.shape
                S = np.vstack((np.zeros(b), self.S))
                S = np.hstack((np.zeros((a+1, 1)), S))
                while t < self.tf:
                    trate=self.get_P(x[1:],t)
                    rate = np.concatenate((np.array([self.fast_rxn]),trate))
                    rate=np.cumsum(rate)

                    t=(t-np.log(np.random.rand(1))/rate[-1])
                    ro=rate[-1]*np.random.rand()

                    while t>self.time[ip]:
                        if t>self.tf:
                            b = len(self.time[ip:])
                            fill = np.repeat(x[1:],b)
                            data[:,ip:]=fill.reshape(__n-1,b)
                            return data
                        else:
                            #data[:,ip]=x.reshape(__n)
                            data[:,ip]=x[1:]
                            ip=ip+1
                    for i in range(len(rate)):
                        if rate[i]>=ro:
                            event=i

                            break

                    x=x+S[:,event].ravel()

            else:   #if not time variant ignore fast reaction

                x = self.xi.flatten()

                #x = self.xi
                __n=len(x)
                self.time=self.gettvec()

                while t<self.tf:
                    rate=self.get_P(x,t)

                    rate=np.cumsum(rate)

                    t=(t-np.log(np.random.rand(1))/rate[-1])
                    ro=rate[-1]*np.random.rand()

                    while t>self.time[ip]:
                        if t>self.tf:
                            b = len(self.time[ip:])
                            fill = np.repeat(x,b)
                            data[:,ip:]=fill.reshape(__n,b)
                            return data
                        else:
                            #data[:,ip]=x.reshape(__n)
                            data[:,ip]=x
                            ip=ip+1
                    for i in range(len(rate)):
                        if rate[i]>=ro:
                            event=i

                            break

                    x=x+self.S[:,event].ravel()



        else:
            'Error'
        self.data=data
        return data

    def _solve(self,n):
        __data=np.zeros((len(self.xi),self.ptimes,n))

        for i in range(n):
            __d=self._run_trajectory()
            __data[:,:,i]=__d
        self.data = __data

        return __data

    def setpar(self,key,val):
        self.params[key]=val

    def get_dist(self,specID=0):
        '''
        build distribution (non-normalized and pdf)
        of rna for the model)
        '''
        n_specs, n_times, n_traj = self.data.shape
        n_specs = int(n_specs)
        n_times = int(n_times)
        n_traj = int(n_traj)
        specID = int(specID)
        max_rna = int(np.max(self.data[specID,:,:]))
        self.pdf = np.zeros((n_times,max_rna+1))
        self.fdist = np.zeros((n_times,max_rna+1))
        for i in range(n_times):
            ind = int(i)
            for j in range(n_traj):
                jnd = int(j)
                self.fdist[ind,int(self.data[int(specID),ind,jnd])] +=1
            self.pdf[ind,:] = self.fdist[ind,:] / np.sum(self.fdist[ind,:])

    def get_traj(self,specID=0,ntraj='all'):
        n_specs, n_times, n_traj = self.data.shape
        n_specs = int(n_specs)
        n_times = int(n_times)




        if isinstance(specID,int):
            if ntraj == 'all':
                return self.data[specID],ntraj
            else:
                try:
                    ntraj = ntraj.flatten().astype(int).tolist()
                except:
                    ntraj = int(ntraj)
                    pass

                return self.data[specID][:,ntraj],ntraj
        else:
            if specID == 'all':
                if ntraj == 'all':
                    return self.data,ntraj
                else:

                    try:
                        ntraj = ntraj.flatten().astype(int).tolist()
                    except:
                        pass

                    return self.data,ntraj

    def get_means(self,specID=0):
        '''
        get the first moment.
        '''
        n_specs, n_times, n_traj = self.data.shape
        max_rna = int(np.max(self.data[int(specID),:,:])+1)
        self.means = np.zeros(n_times)
        for i in range(n_times):
            self.means[i] = np.sum(np.arange(max_rna)*self.pdf[i,:])

    def get_variances(self,specID=0):
        '''
        get the second moment.
        '''
        self.get_dist()
        n_specs, n_times, n_traj = self.data.shape
        max_rna = int(np.max(self.data[specID,:,:])+1)
        self.variances = np.zeros(n_times)
        self.covariances = np.zeros((n_specs,n_specs,n_times))
        for i in range(n_times):
            self.variances[i] = np.sum((np.arange(max_rna)**2)*self.pdf[i,:])-(np.sum(np.arange(max_rna)*self.pdf[i,:])**2)
            self.covariances[:,:,i] = np.cov(self.data[:,i,:])

    def return_all_var(self):
        all_members = self.__dict__.keys()

        return [ (item, self.__dict__[item]) for item in all_members if not item.startswith("_")]

    def return_names(self):
        all_members = self.__dict__.keys()
        return [ item for item in all_members if not item.startswith("_")]






# -*- coding: utf-8 -*-
"""
Created on Thu Dec 17 11:35:43 2020

@author: willi
"""
import re
import time
import os
import numpy as np
from . import CodonDictionaries
from . import FileParser
from . import poi as POI

try:
    from Bio import SeqIO
    from Bio import Entrez
except:

    print('BioPython is not installed, polling genbank will not be possible')
    pass




class Error(Exception):
    """Base class for exceptions in this module."""
    pass


class PathDoesNotExistError(Error):
    """Exception raised for when trying to save a GB file to a directory 
    that doesnt exist

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message
        
class AscNumDoesNotExistError(Error):
    """Exception raised for when trying to pull a gb from an ascession number
    that doesnt exist

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message
        
        
        

class SequenceManipMethods():
    '''
    class that handles manipulation methods dealing with sequences
    '''
    def __init__(self, sequence=''):
        self.sequence = sequence
        self.codon_dicts = CodonDictionaries.CodonDictionaries()
        #get the codon dictionaries


    def optimize_ntseq(self, nt_seq, opt_dict=None):
        '''
        Optimizes a nucleotide sequence

        Parameters
        ----------
        nt_seq : str
            nucleotide sequence string
        opt_dict : dictionary, optional
            a user defined dictionary to optimize over. The default is None.

        Returns
        -------
        opt_seq : str
            Optimized NT sequenced based on a given dictionary of rates

        '''

        if opt_dict is None:
            opt_dict = self.codon_dicts.human_codon_frequency_bias_nakamura

        codons = nt_seq.upper()
        seperated_codons = [codons[i:i+3] for i in range(0, len(codons), 3)] #split codons by 3
        aa = [self.codon_dicts.aa_table[x] for x in seperated_codons]
        opt_seq = ''
        for i in range(0, len(aa)):
            ind = np.argmax([opt_dict[x] for x in self.codon_dicts.aa_table_r[aa[i]]])
            opt_codon = self.codon_dicts.aa_table_r[aa[i]][ind]
            opt_seq = opt_seq + opt_codon
        return opt_seq


    def deoptimize_ntseq(self, nt_seq, deopt_dict=None):

        '''
        Optimizes a nucleotide sequence

        Parameters
        ----------
        nt_seq : str
            nucleotide sequence string
        deopt_dict : dictionary, optional
            a user defined dictionary to deoptimize over. The default is None.

        Returns
        -------
        deopt_seq : str
            Deoptimized NT sequenced based on a given dictionary of rates

        '''

        if deopt_dict is None:
            deopt_dict = self.codon_dicts.human_codon_frequency_bias_nakamura
        codons = nt_seq.upper()
        seperated_codons = [codons[i:i+3] for i in range(0, len(codons), 3)] #split codons by 3
        aa = [self.codon_dicts.aa_table[x] for x in seperated_codons]
        opt_seq = ''
        for i in range(0, len(aa)):
            ind = np.argmin([deopt_dict[x] for x in self.codon_dicts.aa_table_r[aa[i]]])
            opt_codon = self.codon_dicts.aa_table_r[aa[i]][ind]
            opt_seq = opt_seq + opt_codon
        return opt_seq


    def nt2aa(self, nt_seq):
        '''
        Parameters
        ----------
        nt_seq : str
            nucleotide sequence.

        Returns
        -------
        aa : str
            amino acid sequence.

        '''
        aa = ''
        nt_seq = nt_seq.upper()
        for i in range(0, len(nt_seq), 3):
            aa += self.codon_dicts.aa_table[nt_seq[i:i+3]]
        return aa


    def get_gb_file(self, accession_number, save_dir = '.'):
        '''
        A function to poll genbank given an accession number and pull the
        relevant gb file

        *args*

            **accession_number**, the accession number of the sequence to find.
            http://www.nslc.wustl.edu/elgin/genomics/bio4342/1archives/2006/AccReference.pdf

        *keyword args*

            **savetofile**, true or false to save the gb file in the same
            directory as sms for future use



        '''
        
        if not os.path.isdir(save_dir):
            msg = 'Specified save path does not exist, double check the path'\
            ' specified.'
            raise PathDoesNotExistError(msg)

        Entrez.email = "wsraymon@rams.colostate.edu"
        Entrez.tool = 'SingleMoleculeSimulator'
        er = False
        try:
            handle = Entrez.efetch(db="nucleotide", rettype="gb",
                                   retmode="text", id=accession_number)
            #using "gb" as an alias for "genbank"
            gb_record = SeqIO.read(handle, "genbank")
            handle.close()
        except:
            er = True
        time.sleep(2)
        if er == True:
            msg = 'Cannot find given ascession number for genbank, file re'\
                'quest failed.'
            raise AscNumDoesNotExistError(msg)
            


        gb_rec = gb_record
        #gb_obj = gb_record

        #sequence_str = str(gb_record.seq)
        sequence_name = gb_record.name
        
        filename = os.path.join(save_dir, sequence_name+ '.gb')
        f = open(filename, 'w')


        f.write(gb_rec.format('gb'))

        f.close()


    def get_orfs(self, nt_seq='', min_codons=80):

        '''
        Returns open reading frames of the nucleotide sequence given

        orfs = {'1':[proteins],
                '2':[proteins],
                '3':[proteins]}

        *keyword args*

            **nt_seq**, nucleotide sequence as a string. If left blank uses
            the self.sequence_str

            **min_codons**, minimum amount of codons to be considered
            a protein in the open reading frame
        '''

        if nt_seq == '':
            nt_seq = self.sequence.upper()
        nt_seq = nt_seq.upper()
        allstarts = np.array([m.start() for m in re.finditer(
            '(?=A[TU]G((?:.{3})+?)[TU](?:AG|AA|GA))', nt_seq)])


        #allsegments = re.findall('(?=A[TU]G((?:.{3})+?)[TU](?:AG|AA|GA))',self.sequence_str)
        allstops = np.array(
            [m.start() for m in re.finditer('(?=[TU](?:AG|AA|GA))', nt_seq)])
        start_frames = allstarts%3
        stop_frames = allstops%3
        min_len = min_codons*3
        orf1_starts = allstarts[np.where(start_frames == 0)]
        orf2_starts = allstarts[np.where(start_frames == 1)]
        orf3_starts = allstarts[np.where(start_frames == 2)]

        orf1_stops = allstops[np.where(stop_frames == 0)]
        orf2_stops = allstops[np.where(stop_frames == 1)]
        orf3_stops = allstops[np.where(stop_frames == 2)]

        #self.starts = [orf1_starts, orf2_starts, orf3_starts]
        #self.stops = [orf1_stops, orf2_stops, orf3_stops]
        orfs = {'1':[], '2':[], '3':[]}


        orfs = {'1':[], '2':[], '3':[]}
        laststop = 0
        for start in orf1_starts:
            nextstop = orf1_stops[np.where(orf1_stops > start)[0][0]]+3
            if (nextstop - start) > min_len:
                if nextstop != laststop:
                    orfs['1'].append((start, nextstop))

                    laststop = nextstop

        laststop = 0
        for start in orf2_starts:
            nextstop = orf2_stops[np.where(orf2_stops > start)[0][0]]+3
            if (nextstop - start) > min_len:
                if nextstop != laststop:
                    orfs['2'].append((start, nextstop))
                    laststop = nextstop

        laststop = 0
        for start in orf3_starts:
            nextstop = orf3_stops[np.where(orf3_stops > start)[0][0]]+3

            if (nextstop - start) > min_len:
                if nextstop != laststop:
                    orfs['3'].append((start, nextstop))
                    laststop = nextstop



        return orfs
    
    def get_codon_list(self, nt_seq):
        '''
        

        Parameters
        ----------
        nt_seq : str
            nucleotide sequence.

        Returns
        -------
        List of codons.

        '''
        #split codons by 3
        codons = nt_seq.upper()
        seperated_codons = [codons[i:i+3] for i in range(0, len(codons), 3)]
        return seperated_codons
    
    def get_codon_count_dict(self, nt_seq):
        '''
        
        
        Parameters
        ----------
        nt_seq : str
            nucleotide sequence.

        Returns
        -------
        dict of codon counts.

        '''
        #split codons by 3
        codons = nt_seq.upper()
        seperated_codons = [codons[i:i+3] for i in range(0, len(codons), 3)]
        cdict = {}
        for i in range(len(seperated_codons)):
            if seperated_codons[i] in cdict.keys():
                cdict[seperated_codons[i]] +=1
            else:
                cdict[seperated_codons[i]] = 1
            
        return cdict

    def codon_usage(self, nt_seq, codon_dict=None):
        '''
        Analyzes codon useage from the nucleotide sequence

        *args*

            **nt_seq**,  nucleotide sequence as a string

        *returns*

            **codon_sensitivity**, a list of codon sensitivity for the nucleotide sequence

            **cai**, cai value

        '''

        if codon_dict == None:
            codon_dict = self.codon_dicts.human_codon_frequency_bias_nakamura

        codon_usage = np.zeros((1, 21))
        gene_len = len(nt_seq)/3
        aa_seq = self.nt2aa(nt_seq)

        for i in range(len(self.codon_dicts.aa_keys)-1):

            codon_usage[0, i] = len(
                re.findall(self.codon_dicts.aa_keys[i], aa_seq))
        codon_usage[0, 20] = len(re.findall('\*', aa_seq))
        codon_norm = codon_usage/gene_len
        codon_sensitivity = np.round(
            codon_norm*self.codon_dicts.sensitivity_fast_slow, 2)

        cai_codons = []
        for i in range(0, len(nt_seq), 3):
            synonmous_codons = self.codon_dicts.aa_table_r[
                self.codon_dicts.aa_table[codon_dict[nt_seq[i:i+3]]]]

            max_freq = max([codon_dict[x] for x in synonmous_codons])

            cai_codons.append(codon_dict[nt_seq[i:i+3]] /max_freq)

        cai = self.geomean(cai_codons)

        return codon_sensitivity, cai, cai_codons



    def get_proteins(self, orfs, seq, add_tag=True):
        '''
        Parameters
        ----------
        orfs : dict
            dictionary of open reading frames.
            {'1': [[starts],[stops] ],'2': [[starts],[stops] ],'3': [[starts],[stops] ] }
        seq : str
            nucleotide sequence.

        Returns
        -------
        proteins_strs : dict
            aa strings of all proteins found in the given orfs.
        protein_objs : dict
            container objects for proteins found in the given orf.
        proteins_w_tags : dict
            conatiner objects for any proteins with detected tags.

        '''
        cd = self.codon_dicts
        proteins_strs = {'1':[], '2':[], '3':[]}
        protein_objs = {'1':[], '2':[], '3':[]}
        proteins_w_tags = {'1':[], '2':[], '3':[]}


        #tagged_proteins = {a:[] for a in cd.tag_dict.keys()}
        #tagged_protein_seq = {a:[] for a in cd.tag_dict.keys()}

        for i in range(len(orfs)):
            for j in range(len(orfs[str(i+1)])):

                protein = POI.poi()

                pro = self.nt2aa(seq[orfs[str(i+1)][j][0]:orfs[str(i+1)][j][1]])
                nt_seq = seq[orfs[str(i+1)][j][0]:orfs[str(i+1)][j][1]]
                # if pro[-1] == '*':
                #     pro = pro[:-1]
                #     nt_seq = nt_seq[:-3]





                protein.aa_seq = pro
                protein.nt_seq = nt_seq

                proteins_strs[str(i+1)].append(pro)

                protein.gene_length = len(pro) #length of the gene
                protein.tag_length = 0   #length of the tags
                protein.total_length = len(pro)  #total length of the full amino acid sequence
                protein.source_seq = seq
                protein.orf = i
                protein.loc = (orfs[str(i+1)][j][0], orfs[str(i+1)][j][1]+3)
                protein.tags = []

                protein_objs[str(i+1)].append(protein)




        for i in range(len(orfs)):
            for pr in protein_objs[str(i+1)]:

                tag_detected = False

                for tag in cd.tag_dict.keys():

                    if cd.tag_dict[tag] in pr.aa_seq:
                        tag_detected = True


                if tag_detected:
                    self.analyze_protein(pr)
                    pr.tag_added = False
                    proteins_w_tags[str(i+1)].append(pr)
                else:
                    if add_tag:
                        self.add_tag_to_protein(pr)
                        pr.tag_added = True
                    else:
                        pr.tag_added = False
                        self.analyze_protein(pr)



        return proteins_strs, protein_objs, proteins_w_tags

    def add_tag_to_protein(self, POI, tag_type='T_Flag'):
        '''
        Parameters
        ----------
        POI : poi object
            protein of interest object.
        tag_type : str, optional
            What kind of tag to append onto the protein object. The default is 'T_Flag'.

        Returns
        -------
        None.

        '''
        cd = self.codon_dicts

        POI.nt_seq = cd.tag_full[tag_type] + POI.nt_seq
        POI.aa_seq = self.nt2aa(POI.nt_seq)

        self.analyze_protein(POI)





    def analyze_protein(self, poi_obj, epitope_loc='front',):
        cd = self.codon_dicts
        nt_seq = poi_obj.nt_seq
        aa_seq = poi_obj.aa_seq
        #self.POI.name = self.sequence_name
        total_length = len(poi_obj.aa_seq)

        '''
        for key in self.tagged_proteins:
            if protein in self.tagged_proteins[key]:
                self.POI.tag_types.append(key)
        '''
        poi_obj.tag_types = []
        for tag in cd.tag_dict.keys():
            if cd.tag_dict[tag] in aa_seq:
                poi_obj.tag_types.append(tag)

                #''.join(sms.poi[0].split('DYKDDDDK')

        poi_obj.tag_epitopes = {a:[] for a in poi_obj.tag_types}
        gs = poi_obj.aa_seq


        for i in range(len(poi_obj.tag_types)):

            try:
                nt_tag = cd.tag_full[poi_obj.tag_types[i]]
                aa_tag = self.nt2aa(nt_tag)
            except:
                epi = cd.tag_dict[poi_obj.tag_types[i]]
                firstep = poi_obj.aa_seq.find(epi)
                lastep = len(poi_obj.aa_seq) - poi_obj.aa_seq[::-1].find(epi[::-1])
                aa_tag = poi_obj.aa_seq[firstep:lastep]
                nt_tag = poi_obj.nt_seq[3*firstep:3*lastep]

            if epitope_loc == 'front':
                offset = 0
            if epitope_loc == 'middle':
                offset = int(len(cd.tag_dict[poi_obj.tag_types[i]])/2)
            if epitope_loc == 'back':
                offset = len(cd.tag_dict[poi_obj.tag_types[i]])

            poi_obj.tag_epitopes[poi_obj.tag_types[i]] = [
                m.start()+1+offset for m in re.finditer(
                    cd.tag_dict[poi_obj.tag_types[i]], poi_obj.aa_seq)]

            gs = gs.replace(aa_tag, '')
            
        poi_obj.gene_seq = gs
        poi_obj.gene_length = len(gs)
        poi_obj.total_length = total_length
        
        taglocs = np.array([x for x in poi_obj.tag_epitopes.values()])
        if taglocs.shape[0] > 0:
            tag_start,tag_stop = (int(np.min(taglocs)), int(np.max(taglocs)))
        
            poi_obj.tag_seq = aa_seq[tag_start:tag_stop]
            poi_obj.tag_length = len(aa_seq[tag_start:tag_stop])
        else:
            poi_obj.tag_seq = ''
            poi_obj.tag_length = 0

        codons = []
        for i in range(0, len(nt_seq), 3):
            codons.append(nt_seq[i:i+3])
        #POI.codons = codons

        #POI.codon_sensitivity, POI.CAI, POI.CAI_codons = self.codon_usage(POI.nt_seq)
        poi_obj.ki = .03
        poi_obj.ke = 10
        poi_obj.kt = 10

    def seq_to_protein_obj(self, sequence_str, min_codons=80, add_tag=True):

        orfs = self.get_orfs(sequence_str, min_codons=min_codons)
        _, proteins, _ = self.get_proteins(
            orfs, sequence_str, add_tag=add_tag)

        return proteins


    def get_largest_poi(self,seqfile, min_codons=80, add_tag=True):
        '''
        Convenience file to get the largest poi if you know your file 
        has multiple orfs

        Parameters
        ----------
        seqfile : sequence file
            a file containting sequence data to get a mRNA sequence out of.
        min_codons : iny, optional
            minimum contingous codons to consider an ORF. The default is 80.
        add_tag : bool, optional
            Add a fluorescent tag if none is found. The default is True.

        Returns
        -------
        POI obj.

        '''
        fp = FileParser.FileParser()
        sequence_str = fp.get_sequence(seqfile).upper()
        orfs = self.get_orfs(sequence_str, min_codons=min_codons)
        protein_strs, proteins, tagged_proteins = self.get_proteins(orfs,
                                                                    sequence_str,
                                                                    add_tag=add_tag)
        
        sizes = [[len(y) for y in x] for x in protein_strs.values()]
        maxsize = max([item for sublist in sizes for item in sublist])

        for i in range(3):
            if maxsize in sizes[i]:
                frame = i
                pindex = sizes[i].index(maxsize)
                
        return proteins[str(int(i))][pindex]


    def open_seq_file(self, seqfile, min_codons=80, add_tag=True):

        '''
        Reads a sequence file, either a .txt file or a .gb genbank file

        *args*

            **seqfile**, sequence file either in txt, gb, gbk format
        '''

        fp = FileParser.FileParser()
        #TODO expose this to the user:
        #sequence_name = fp.get_name(seqfile)
        #sequence_description = fp.get_description(seqfile)
        
        sequence_str = fp.get_sequence(seqfile)
        if isinstance(sequence_str, str):
            sequence_str = sequence_str.upper()
            orfs = self.get_orfs(sequence_str, min_codons=min_codons)
    
            protein_strs, proteins, tagged_proteins = self.get_proteins(orfs,
                                                                        sequence_str,
                                                                        add_tag=add_tag)
        if isinstance(sequence_str, list):
            
            #multiple sequence fasta detected, split them up and handle them 
            #seperately, then merge them into the dictionaries later
            ps_list = []
            pro_list = []
            tagd_list = []
            for sequence in sequence_str:
                sequence = sequence.upper()
                orfs = self.get_orfs(sequence, min_codons=min_codons)
                
                protein_strs, proteins, tagged_proteins = self.get_proteins(orfs,
                                                                            sequence,
                                                                            add_tag=add_tag)
                ps_list.append(protein_strs)
                pro_list.append(proteins)
                tagd_list.append(tagged_proteins)
                
            # compress the dictionaries by keys
            flatten_dict = lambda dictionary: [[item for sublist in [y[x] for y in dictionary]
                                                for item in sublist] for x in ['1','2','3']]
            

            protein_strs = {i: flatten_dict(ps_list)[int(i)-1] for i in ['1','2','3']}
            proteins = {i: flatten_dict(pro_list)[int(i)-1] for i in ['1','2','3']}
            tagged_proteins = {i: flatten_dict(tagd_list)[int(i)-1] for i in ['1','2','3']}
                
        return protein_strs, proteins, tagged_proteins, sequence_str



    def get_tag_loc(self, aa_seq, tag, epitope_loc='front'):
        cd = self.codon_dicts

        if epitope_loc == 'front':
            offset = 0
        if epitope_loc == 'middle':
            offset = int(len(tag)/2)
        if epitope_loc == 'back':
            offset = len(tag)

        return [m.start()+1+offset for m in re.finditer(tag, aa_seq)]


    @staticmethod
    def geomean(iterable):
        '''geometric mean used for codon sensitivity calculations
        '''
        a = np.array(iterable)
        return a.prod()**(1.0/len(a))



# -*- coding: utf-8 -*-
"""
Created on Thu Dec 17 11:50:56 2020

@author: William Raymond
"""
import numpy as np

class Error(Exception):
    """Base class for exceptions in this module."""
    pass

class UnrecognizedNormalizationError(Error):
    """Exception raised for errors in the input.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class IntensityAnalyses():
    def __init__(self):
        pass

    @staticmethod
    def get_acc2(data, trunc=False):
        '''
        return autocorrelation from a data vector via Fourier transform

        Parameters
        ----------
        data : ndarray
            numpy array of data to get the correlation from.
        trunc : bool, optional
            Remove non zero entries and only return decreasing entries.
            The default is False.

        Returns
        -------
        autocorrelation : ndarray
            autocorrelation function.

        '''

        N = len(data)
        fvi = np.fft.fft(data, n=2*N)
        acf = fvi*np.conjugate(fvi)
        acf = np.fft.ifft(acf)
        acf = np.real(acf[:N])/float(N)
        if trunc:
            acf[acf < 0] = 0
            for i in range(1, len(acf)):
                if acf[i] > acf[i-1]:
                    acf[i] = acf[i-1]
        return acf


    def get_fragments(self, position_tensor, total_length=None):
        #TODO
        '''
        ..TODO refactor this code!!
        
        Get individual ribosome trajectories for kymograph generation

        .. warning:: This is not perfect process, if the time resolution of a
        simulation is too coarse this will not be able to rebuild ribosome
        trajectories. This will fail if ribosomes are moving fast enough 
        between timesteps to pass previously recorded positions.

        Parameters
        ----------
        position_tensor : ndarray
            Ribosome position tensor:  Nribosomes x Ntime, accessed from
            SSA_Soln.solutions if a full statistical model was run.
        total_length : int, optional
            Length of the transcript. If left blank it will use the maximum
            detected ribosome position

        Returns
        -------
        fragtimes : ndarray
            N_ribosomes x N_longest_time. The individual trajectory times
            for the positions fragments
        fragarray : ndarray
            N_ribosomes x N_position. The individual ribosome trajectory
            positions

        '''
        if total_length == None:
            total_length = np.max(position_tensor)

        fragmented_trajectories = []
        fragtimes = []
        maxlen = 0

        fragmentspertraj = []
        solutions = [position_tensor]
        nsteps = position_tensor.shape[0]
        k = 0


        ind = np.zeros(position_tensor.shape[1]).astype(int)
        for i in range(position_tensor.shape[1]):
          tmp = np.where(position_tensor[:, i] == 0)[0]
          if len(tmp) > 0:
            ind[i] = np.where(position_tensor[:, i] == 0)[0][0]
          else:
            ind[i] = position_tensor.shape[0]-1

        changes = ind[1:] - ind[:-1]
        addindexes = np.where(changes > 0)[0]
        subindexes = np.where(changes < 0)[0]

        sub = solutions[k][:, 1:] - solutions[k][:, :-1]
        neutralindexes = np.unique(np.where(sub < 0)[1])
        neutralindexes = np.setxor1d(neutralindexes, subindexes)

        for index in neutralindexes:
            pre = solutions[k][:, index]
            post = solutions[k][:, index+1]
            changecount = 0
            while len(np.where(post - pre < 0)[0]) > 0:

                post = np.append([total_length], post)
                pre = np.append(pre, 0)

                changecount += 1

            for i in range(changecount):
                addindexes = np.sort(np.append(addindexes, index))
                subindexes = np.sort(np.append(subindexes, index))

            changes[index] = -changecount
            ind[index] += changecount


        for index in np.where(np.abs(changes) > 1)[0]:
            if changes[index] < 0:
                for i in range(np.abs(changes[index])-1):
                    subindexes = np.sort(np.append(subindexes, index))
            else:
                for i in range(np.abs(changes[index])-1):
                    addindexes = np.sort(np.append(addindexes, index))

        truefrags = len(subindexes)




        if len(subindexes) < len(addindexes):
            subindexes = np.append(
                subindexes, (
                    np.ones((
                        len(addindexes) - len(subindexes)))*(nsteps-1)).astype(int))


        fragmentspertraj.append(len(subindexes))

        for m in range(min(len(subindexes), len(addindexes))):
            traj = solutions[k][:, addindexes[m]:subindexes[m]+1]
            traj_ind = changes[addindexes[m]:subindexes[m]+1]

            startind = ind[addindexes[m]]
            minusloc = [0] + np.where(traj_ind < 0)[0].astype(int).tolist()
            fragment = np.array([])



            iterind = startind

            if subindexes[m]-addindexes[m] > 0:
                if len(minusloc) > 1:
                    if m <= truefrags:
                        for n in range(len(minusloc)-1):
                            iterind = iterind + int(min(0, traj_ind[minusloc[n]]))

                            fragment = np.append(fragment, traj[iterind, minusloc[n]+1:minusloc[n+1]+1].flatten())

                        fragment = np.append(fragment, traj[0, minusloc[-1]+1:].flatten())

                    else:
                        for n in range(len(minusloc)-1):

                            iterind = iterind + min(0, traj_ind[minusloc[n]])

                            fragment = np.append(fragment, traj[iterind, minusloc[n]+1:minusloc[n+1]+1].flatten())


                        fragment = np.append(fragment, traj[m-truefrags, minusloc[-1]+1:].flatten())



                else:

                    fragment = solutions[k][startind][addindexes[m]:subindexes[m]+1].flatten()



                fragtimes.append(addindexes[m]+1)


                fragmented_trajectories.append(fragment)
                #if m <= truefrags:
                    #kes.append(genelength/truetime[len(fragment)])

                if len(fragment) > maxlen:
                    maxlen = len(fragment)


        fragarray = np.zeros((len(fragmented_trajectories), maxlen))
        for i in range(len(fragmented_trajectories)):
            fragarray[i][0:len(fragmented_trajectories[i])] = fragmented_trajectories[i]

        return fragtimes, fragarray



    # def get_frags_ssa_obj(self,ssa_obj, total_length = None):

    #     fragmented_trajectories = []
    #     fragtimes = []
    #     maxlen = 0

    #     fragmentspertraj= []
    #     n_traj = ssa_obj.n_traj

    #     solutions = ssa_obj.solutions
    #     if total_length == None:
    #         total_length = 0
    #         for k in range(n_traj):
    #             if total_length < np.max(solutions[k]):
    #                 total_length = np.max(solutions[k])

    #     nsteps = solutions[0].shape[1]
    #     for k in range(n_traj):
    #         ind = np.array([next(j for j in range(0,solutions[k].shape[0]) if int(solutions[k][j, i]) == 0 or int(solutions[k][j, i]) == -1) for i in range(0, solutions[k].shape[1])])
    #         changes = ind[1:] - ind[:-1]
    #         addindexes = np.where(changes > 0)[0]
    #         subindexes = np.where(changes < 0)[0]

    #         sub = solutions[k][:,1:] - solutions[k][:,:-1]
    #         neutralindexes = np.unique(np.where(sub < 0)[1])
    #         neutralindexes = np.setxor1d(neutralindexes, subindexes)

    #         for index in neutralindexes:
    #             pre = solutions[k][:,index]
    #             post = solutions[k][:,index+1]
    #             changecount = 0
    #             while len(np.where(post - pre < 0)[0]) > 0:

    #                 post = np.append([total_length],post)
    #                 pre = np.append(pre,0)

    #                 changecount+=1

    #             for i in range(changecount):
    #                 addindexes = np.sort(np.append(addindexes,index))
    #                 subindexes = np.sort(np.append(subindexes,index))

    #             changes[index] = -changecount
    #             ind[index] += changecount


    #         for index in np.where(np.abs(changes)>1)[0]:
    #             if changes[index] < 0:
    #                 for i in range(np.abs(changes[index])-1):
    #                     subindexes = np.sort(np.append(subindexes,index))
    #             else:
    #                 for i in range(np.abs(changes[index])-1):
    #                     addindexes = np.sort(np.append(addindexes,index))

    #         truefrags = len(subindexes)




    #         if len(subindexes) < len(addindexes):
    #             subindexes = np.append(subindexes, (np.ones((len(addindexes)-len(subindexes)))*(nsteps-1)).astype(int))


    #         fragmentspertraj.append(len(subindexes))

    #         for m in range(min(len(subindexes),len(addindexes))):
    #             traj = solutions[k][:, addindexes[m]:subindexes[m]+1]
    #             traj_ind = changes[addindexes[m]:subindexes[m]+1]

    #             startind = ind[addindexes[m]]
    #             minusloc = [0] + np.where(traj_ind < 0)[0].astype(int).tolist()
    #             fragment = np.array([])



    #             iterind = startind

    #             if subindexes[m]-addindexes[m] > 0:
    #                 if len(minusloc) > 1:
    #                     if m <= truefrags:
    #                         for n in range(len(minusloc)-1):
    #                             iterind = iterind + min(0,traj_ind[minusloc[n]])
    #                             fragment = np.append(fragment, traj[iterind, minusloc[n]+1:minusloc[n+1]+1].flatten())







    #                         fragment = np.append(fragment, traj[0, minusloc[-1]+1:].flatten())

    #                     else:
    #                         for n in range(len(minusloc)-1):

    #                             iterind = iterind + min(0,traj_ind[minusloc[n]])

    #                             fragment = np.append(fragment, traj[iterind, minusloc[n]+1:minusloc[n+1]+1].flatten())


    #                         fragment = np.append(fragment, traj[m-truefrags, minusloc[-1]+1:].flatten())



    #                 else:

    #                     fragment = solutions[k][startind][addindexes[m]:subindexes[m]+1].flatten()



    #                 fragtimes.append(addindexes[m]+1)


    #                 fragmented_trajectories.append(fragment)
    #                 #if m <= truefrags:
    #                     #kes.append(genelength/truetime[len(fragment)])

    #                 if len(fragment) > maxlen:
    #                     maxlen = len(fragment)


    #         fragarray = np.zeros((len(fragmented_trajectories), maxlen))
    #         for i in range(len(fragmented_trajectories)):
    #             fragarray[i][0:len(fragmented_trajectories[i])] = fragmented_trajectories[i]


    #     ssa_obj.fragments = fragarray
    #     ssa_obj.fragtimes = fragtimes
    #     ssa_obj.frag_per_traj = fragmentspertraj
    #     ssa_obj.full_frags = truefrags


    ##TODO Convert excitation of the intensity vector (realistic probe)

    def get_g0(self, correlation, mode='interp'):
        '''


        Parameters
        ----------
        correlation : ndarray
            numpy array of a cross or autocovariance.
        mode : string, optional
            the type of G0 shot noise to return,

            * 'Interp' - will interpolate the g0 position from the G1, G2, and G3 points.
            * 'g1' - (second point) g1 will be returned
            * 'g0' - g0 will be returned (first point)
            * 'max' -maximum of the correlation will be returned
            The default is 'interp'.

        Returns
        -------
        G0 : float
            point to normalize correlation over.

        '''
        if mode.lower() in ['interp', 'inter', 'extrapolate', 'interpolate']:
            X = [1, 2, 3, 4]
            V = correlation[:, X, :]
            G0 = np.interp(0, X, V)

        if mode.lower() in ['g1', '1']:
            G0 = correlation[:, 1, :]

        if mode.lower() in ['g0', '0']:
            G0 = correlation[:, 0, :]

        if mode.lower() in ['max', 'maximum']:
            G0 = np.max(correlation, axis=1)
        return G0

    def normalize_cc(self, correlation, mode='global_max'):
        '''
        normalize a cross correlation

        Parameters
        ----------
        correlation : ndarray
            numpy array of the cross correlation to .
        mode : str, optional
            The type of normalization. The default is 'max'.

            * 'max' - normalize by maximum point
            * 'middle' - maximize by the middle point

        Returns
        -------
        norm_cor : ndarray
            normalized cross correlation.

        '''
        if mode.lower() in ['global_max', 'global_maximum', 'gmax']:
            norm_cor = correlation/np.max(correlation)

        if mode.lower() in ['indiv_max', 'indiv_maximum', 'imax',
                            'individual_maximum', 'individual_max']:
            norm_cor = correlation/np.max(correlation, 1)[:, np.newaxis, :]

        if mode.lower() in ['indiv_center', 'indiv_middle',
                            'individual_middle',
                            'individual_center']:
            centerpoint = int((correlation.shape[1]+1)/2)-1
            norm_cor = correlation/(
                correlation[:, centerpoint][:, np.newaxis, :])

        if mode.lower() in ['global_center', 'global_middle']:
            centerpoint = int((correlation.shape[1]+1)/2)-1
            norm_cor = correlation/(np.mean(correlation[:, centerpoint]))
        if mode.lower() in ['raw']:
            norm_cor = correlation

        return norm_cor


    def get_crosscorr(self, intensity_vecs, norm='indiv', g0='indiv_center'):
        '''
        return a cross correlation tensor from an intensity tensor of (ncolor, ntime, ntraj)

        Parameters
        ----------
        intensity_vecs : ndarray
            (ncolor, ntime, ntraj) intensity tensor.
        norm : str, optional
            Normalization to apply,
            * global - subtract by the global intensity mean for the correlation
            * indiv - subtract by each trajectory's mean for the correlation
            * raw - do no mean subtraction
            The default is 'indiv'.
        g0 : str, optional
            point to normalize the correlation by after correlating
            * global_middle - divide the correlation by the average of the center point of all trajectories
            * indiv_middle - divide the correlation trajectories by their individual center point
            * global_max - divide the correlation trajectories by the global max point
            * indiv_max - divide the correlation trajectories by their individual max point
            * raw - do not divide by a point.

        Returns
        -------
        cross_corr : ndarray
            cross correlation tensor,  size (ncolor**2, 2*ntime-1, ntraj).
        err_crosscorr : TYPE
            cross correlation standard error estimation.

            .. math:: \sigma_{cc} = \frac{\sigma}{\sqrt(n_{trajectories})}

        inds : list
            indices describing which colors were correlated with which.

        '''

        ncolors = intensity_vecs.shape[0]
        time_pts = intensity_vecs.shape[1]
        traj = intensity_vecs.shape[2]
        cross_corr = np.zeros((ncolors**2, time_pts*2-1, traj))

        i = 0
        k = 0
        inds = []


        for n in range(intensity_vecs.shape[0]):
            for m in range(intensity_vecs.shape[0]):
                iv1 = intensity_vecs[n].T
                iv2 = intensity_vecs[m].T
                inds.append((n, m))

                if norm in ['global', 'Global', 'g', 'G']:
                    global_mean1 = np.mean(iv1)
                    global_mean2 = np.mean(iv2)

                # slen = np.correlate(iv1[i]-np.mean(iv1[i]),iv2[i]-np.mean(iv2[i]),'full').shape[0]
                # crosscorr_vec = np.zeros((iv1.shape[0],slen))

                for i in range(traj):
                    if norm in ['Individual', 'I', 'individual', 'ind', 'indiv']:
                        cross_corr[k, :, i] = np.correlate(
                            iv1[i, :]-np.mean(iv1[i, :]),
                            iv2[i, :]-np.mean(iv2[i, :]), 'full')/time_pts

                    elif norm in ['global', 'Global', 'g', 'G']:
                        cross_corr[k, :, i] = np.correlate(
                            iv1[i, :]-global_mean1, iv2[i, :]-global_mean2, 'full')/time_pts

                    elif norm in ['raw', 'Raw', None, 'none', 'None']:
                        cross_corr[k, :, i] = np.correlate(
                            iv1[i, :], iv2[i, :], 'full')/time_pts

                    else:
                        msg = 'unrecognized normalization, please use '\
                              'individual, global, or none for norm arguement'
                        raise UnrecognizedNormalizationError(msg)
                        

                k += 1
        if g0 != None:
            cross_corr = self.normalize_cc(cross_corr, mode=g0)

        err_crosscorr = 1.0/np.sqrt(traj)*np.std(cross_corr, ddof=1, axis=2)

        return cross_corr, err_crosscorr, inds

    # def get_crosscorr2(self, iv1,iv2):
    #     '''
    #     returns the autocorrelations
    #     '''

    #     i = 0
    #     slen = np.correlate(iv1[i]-np.mean(iv1[i]),iv2[i]-np.mean(iv2[i]),'full').shape[0]
    #     crosscorr_vec = np.zeros((iv1.shape[0],slen))

    #     for i in range(iv1.shape[0]):
    #         crosscorr_vec[i,:] = np.correlate(iv1[i]-np.mean(iv1[i]),iv2[i]-np.mean(iv2[i]),'full')/len(iv1)

    #     normalized_autocorr = crosscorr_vec.T/ crosscorr_vec[:,len(iv1[i])-1]
    #     mean_autocorr = np.mean(normalized_autocorr, axis=1)



    #     return crosscorr_vec, mean_autocorr


    def get_autocorr_norm(self, intensity_vec, time_vec,
                          geneLength, normalization='Individual'):
        '''
        returns the autocorrelations
        '''

        autocorr_vec = np.zeros((intensity_vec.shape))

        if normalization in ['Individual', 'I', 'individual', 'ind']:
            for i in range(intensity_vec.shape[0]):
                autocorr_vec[i, :] = self.get_acc2(intensity_vec[i] - np.mean(intensity_vec[i]))
        elif normalization in ['global', 'Global', 'g', 'G']:
            global_mean = np.mean(intensity_vec)
            for i in range(intensity_vec.shape[0]):
                autocorr_vec[i, :] = self.get_acc2(
                    intensity_vec[i] - global_mean)

        else:
            print('unrecognized normalization, using indivdual means')
            for i in range(intensity_vec.shape[0]):
                autocorr_vec[i, :] = self.get_acc2(
                    intensity_vec[i] - np.mean(intensity_vec[i]))

        normalized_autocorr = autocorr_vec.T / autocorr_vec[:, 0]
        mean_autocorr = np.mean(normalized_autocorr, axis=1)

        error_autocorr = np.std(normalized_autocorr, axis=1)/np.sqrt(intensity_vec.shape[0])

        dwelltime = None

        try:
            dwelltime = time_vec[np.where(mean_autocorr < .01)[0][0]]

        except:
            try:
                dwelltime = time_vec[np.where(mean_autocorr < .05)[0][0]]

            except:
                dwelltime = 1


        try:
            zeroind = np.where(mean_autocorr < 0)[0][0]
            length = int(.3*len(mean_autocorr))
            zeromean = np.mean(mean_autocorr[-length:])
            zeromean2 = np.mean(mean_autocorr[zeroind:])

            normalized_autocorr = normalized_autocorr - zeromean2
            mean_autocorr = np.mean(normalized_autocorr, axis=1)

            error_autocorr = np.std(normalized_autocorr, axis=1)/np.sqrt(intensity_vec.shape[0])

        except:
            pass



        ke_exp = np.round(geneLength/dwelltime, 1)

        return normalized_autocorr, mean_autocorr, error_autocorr, dwelltime, ke_exp



    def get_autocov(self, intensity_vec, norm='raw'):
        autocorr_vec = np.zeros((intensity_vec.shape))
        autocorr_err = np.zeros((intensity_vec.shape))
        colors = intensity_vec.shape[0]
        n_traj = intensity_vec.shape[2]

        for n in range(colors):
            if norm in ['Individual', 'I', 'individual', 'ind']:
                for i in range(intensity_vec.shape[2]):
                    ivec = intensity_vec[n, :, i]
                    autocorr_vec[n, :, i] = self.get_acc2(
                        (ivec - np.mean(ivec))/np.var(ivec))

            elif norm in ['global', 'Global', 'g', 'G']:
                global_mean = np.mean(intensity_vec[n])
                global_var = np.var(intensity_vec[n])
                for i in range(intensity_vec.shape[2]):
                    autocorr_vec[n, :, i] = self.get_acc2(
                        (intensity_vec[n, :, i]-global_mean)/global_var )
            elif norm in ['raw', 'Raw']:
                for i in range(intensity_vec.shape[2]):
                    autocorr_vec[n, :, i] = self.get_acc2(
                        intensity_vec[n, :, i])
            else:
                print('unrecognized normalization,'/
                      ' please use individual, global, or none')
                return

        autocorr_err = 1.0/np.sqrt(n_traj)*np.std(
            autocorr_vec, ddof=1, axis=2)

        return autocorr_vec, autocorr_err


    def get_autocorr(self, autocov, g0='G0'):
        '''
        normalize the autocovariance over g0
        '''
        autocorr = np.copy(autocov)
        n_traj = autocorr.shape[-1]

        g0 = self.get_g0(autocov, g0)
        for n in range(autocov.shape[0]):
            autocorr[n] = autocorr[n]/g0[n]

        err_autocorr =  1.0/np.sqrt(n_traj)*np.std(autocorr, ddof=1, axis=2)
        return autocorr, err_autocorr



    def get_autocorr2(self, intensity_vec, time_vec, totalSimulationTime,
                      geneLength, normalization='Individual'):
        '''
        returns the autocorrelations
        '''

        autocorr_vec = np.zeros((intensity_vec.shape))


        if normalization in ['Individual', 'I', 'individual', 'ind']:
            for i in range(intensity_vec.shape[0]):
                autocorr_vec[i, :] = self.get_acc2(
                    intensity_vec[i]-np.mean(intensity_vec[i]))
        elif normalization in ['global', 'Global', 'g', 'G']:
            global_mean = np.mean(intensity_vec)
            for i in range(intensity_vec.shape[0]):
                autocorr_vec[i, :] = self.get_acc2(
                    intensity_vec[i]-global_mean)

        else:
            print('unrecognized normalization, using indivdual means')
            for i in range(intensity_vec.shape[0]):
                autocorr_vec[i, :] = self.get_acc2(
                    intensity_vec[i]-np.mean(intensity_vec[i]))
        autocorr = autocorr_vec.T
        mean_autocorr = np.mean(autocorr, axis=1)
        error_autocorr = np.std(
            autocorr, axis=1)/np.sqrt(intensity_vec.shape[0])


        dwelltime = None

        try:
            dwelltime = time_vec[np.where(mean_autocorr < .01)[0][0]]

        except:
            try:
                dwelltime = time_vec[np.where(mean_autocorr < .05)[0][0]]
            except:
                dwelltime = 1

        ke_exp = np.round(geneLength/dwelltime, 1)

        return autocorr, mean_autocorr, error_autocorr, dwelltime, ke_exp
